import json
import logging
from abc import abstractmethod
from functools import lru_cache
from pathlib import Path
from typing import Mapping

from evaluate import load as load_metric
from llama_cpp import LlamaGrammar

from evoprompt.helpers.prompts import BasePromptsFromJsonMixin
from evoprompt.models import LLMModel
from evoprompt.task import Task
from evoprompt.task.task import DatasetDatum

logger = logging.getLogger(__name__)


class Simplification(Task):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.metric = load_metric("evaluate-metric/sari")

    def _evaluate_sample(self, prompt: str, datum: DatasetDatum):
        gold_label = self._get_gold_label_for_datum(datum)
        response, usage = self.predict(prompt=prompt, datum=datum)
        response = response.lower()

        scores = self.metric.compute(
            sources=[self._get_text_for_datum(datum)],
            predictions=[response],
            references=[gold_label],
        )
        return scores["sari"], usage

    @lru_cache
    def _get_grammer(self, datum: DatasetDatum, verbose: bool = False):
        return None

    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        return "\nInput: " + '"' + self._get_text_for_datum(datum) + '"'

    @abstractmethod
    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        pass

    def _aggregate_result(self, results: list[str]) -> float:
        return sum(results) / len(results)

    @property
    def metric_name(self):
        return "sari"


class ASSET(BasePromptsFromJsonMixin, Simplification):
    shorthand = "asset"
    base_prompts_file = "evoprompt/initial_prompts/asset/prompts.json"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args,
            validation_dataset="facebook/asset",
            test_dataset="facebook/asset",
            validation_split="validation",
            test_split="test",
            **kwargs
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["original"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["simplifications"]
