import inspect
import logging
import os
import re
from functools import wraps
from pathlib import Path
from pprint import pformat
from textwrap import dedent, indent
from typing import Any, Callable
from uuid import uuid4

import numpy
import numpy.typing

rng = None
global_seed = None

current_directory = Path(__file__).resolve().parent
logger = logging.getLogger(__name__)

# a global file handler to keep track of ongoing run's file logger
file_handler = None


def init_rng(seed: int | None):
    global rng, global_seed
    global_seed = seed
    rng = numpy.random.default_rng(seed)
    logger.info("Initialized rng with seed %s.", seed)


def get_seed() -> int:
    return global_seed


def get_rng() -> numpy.random.Generator:
    if rng is None:
        raise RuntimeError(
            "Rng was not initialized. Make sure to run `%s` first."
            % (init_rng.__module__ + "." + init_rng.__qualname__)
        )
    return rng


def setup_console_logger(verbosity_level: int = 0):
    # create console handler
    console_handler = logging.StreamHandler()
    # we currently only have one verbosity level which is DEBUG
    console_handler.setLevel(logging.DEBUG if verbosity_level > 0 else logging.INFO)
    # formatter for console handler
    console_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s: %(message)s"
    )
    console_handler.setFormatter(console_formatter)
    # configure root logger; level here refers to the lowest level of the logging events the logger accepts
    logging.basicConfig(handlers=(console_handler,), level=logging.NOTSET)


run_name_prompt = """
Create a random name that sounds german or dutch
The parts should be separated by underscores and contain only lowercase.
Only return the name without any text before or after.""".strip()

RUNS_DIR = current_directory.parent / "runs"
RUNS_DIR.mkdir(exist_ok=True)


def initialize_run_directory(model: Callable):
    global file_handler

    # get package logger
    logger = logging.getLogger(__package__)

    # remove file handler of previous run
    if file_handler is not None:
        logger.removeHandler(file_handler)

    response, _, _ = model(None, run_name_prompt)
    run_name_match = re.search(r"^\w+$", response, re.MULTILINE)
    existing_run_names = os.listdir(RUNS_DIR) if RUNS_DIR.exists() else []
    if run_name_match is None or run_name_match.group(0) in existing_run_names:
        run_name = uuid4().hex
    else:
        run_name = run_name_match.group(0)
    run_directory = RUNS_DIR / run_name
    run_directory.mkdir(parents=True, exist_ok=False)
    # create file handler and set level to debug
    file_handler = logging.FileHandler(run_directory / "output.log")
    file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("{asctime} {levelname} {message}", style="{")
    file_handler.setFormatter(formatter)
    # add file handler to logger
    logger.addHandler(file_handler)

    logger.info(f"Hello my name is {run_name} and I live in {run_directory}")
    return run_directory


def escape_whitespace(text: str) -> str:
    return text.replace("\r", "\\r").replace("\n", "\\n")


class log_calls:
    description: str

    prolog = dedent(
        """
    ----------------------------------------
    {description}
    {func_name}:
        \tArguments:
        {arguments}
        """
    )

    epilog = dedent(
        """\tResult:
        {result}
    ----------------------------------------
    """
    )

    def __init__(self, description: str, level: int = logging.DEBUG):
        self.description = description
        self.level = level
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        self.logger = logging.getLogger(mod.__name__)

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            arguments = self._get_named_arguments(func, *args, **kwargs)
            self.logger.log(
                self.level,
                self.prolog.format(
                    description=self.description,
                    func_name=func.__name__,
                    arguments=indent(pformat(arguments), "\t"),
                ),
            )

            result = func(*args, **kwargs)

            self.logger.log(
                self.level,
                self.epilog.format(
                    result=indent(pformat(result), "\t"),
                ),
            )
            return result

        return wrapper

    def _get_named_arguments(self, func: Callable[..., Any], *args, **kwargs):
        signature = inspect.signature(func)
        arguments = {}
        for argument_index, (argument_name, argument) in enumerate(
            signature.parameters.items()
        ):
            if argument_index < len(args):
                # argument_name is from args
                value = args[argument_index]
            elif argument_name in kwargs:
                # argument_name is from kwargs
                value = kwargs[argument_name]
            else:
                # argument_name is from defaults
                value = argument.default

            arguments[argument_name] = value
        return arguments


def get_all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )
